(* olex.mll
 * 
 * Example of using a lexer generator to create a program.
 * 
 * To compile this program to an executable do:
 *  ocamllex olex.mll
 *  ocamlc -o olex olex.ml
 *
 * Ocamllex Tutorial:
 *   http://plus.kaist.ac.kr/~shoh/ocaml/ocamllex-ocamlyacc/ocamllex-tutorial/
 * Lexing:
 *   http://caml.inria.fr/pub/docs/manual-ocaml/libref/Lexing.html
 *)

(* header section *)
{  
  open Lexing
  open String

  let incr_linenum lexbuf =
    let pos = lexbuf.Lexing.lex_curr_p in
    lexbuf.Lexing.lex_curr_p <- { pos with
      Lexing.pos_lnum = pos.Lexing.pos_lnum + 1;
      Lexing.pos_bol = pos.Lexing.pos_cnum;
    }
    
  let unexpected_char c lexbuf =
    Printf.printf "Char %s is unexpected at (line:%d, column:%d).\n" 
                  (Char.escaped c) 
                  (lexeme_start_p lexbuf).pos_lnum
                  ((lexeme_start_p lexbuf).pos_cnum - (lexeme_start_p lexbuf).pos_bol + 1)
        
}

(* definitions section *)
let character = ['a'-'z''A'-'Z']
let digit = ['0'-'9']
let underscore = ['_']
let whitespace = [' ' '\t']
let regname = "eax" | "ebx" | "ecx" | "edx" | "ebp" | "esp" | "epi" | "edi"

(* rules section *)
rule lex = parse
  | "if" 
      { Printf.printf "IF: %s\n" (lexeme lexbuf);
        lex lexbuf
      }
  | character (character|digit|underscore)*
      { Printf.printf "ID: %s\n" (lexeme lexbuf);
        lex lexbuf
      }
  | digit+            
      { Printf.printf "INT: %s\n" (lexeme lexbuf);
        lex lexbuf
      }
  | '(' 
      { Printf.printf "LPAREN: %s\n" (lexeme lexbuf); 
        lex lexbuf 
      }
  | ')' 
      { Printf.printf "RPAREN: %s\n" (lexeme lexbuf); 
        lex lexbuf 
      }
  | whitespace+ 
      { lex lexbuf }
  | '"'[^'"']'"'
      { Printf.printf "STRING: %s\n" (lexeme lexbuf); 
        lex lexbuf
      }
  | '\n' 
      { incr_linenum lexbuf; lex lexbuf }
  | (('-')?(digit+)) as disp 
    '(' '%' (regname as breg) ',' 
        '%' (regname as ireg) ',' 
        (digit+ as scale) 
    ')' 
      {
        Printf.printf "Ind({i_base=Some %s; i_iscl=Some (%s, %sl); i_disp=Some(DImm (Int32.of_int (%sl)))})\n" 
                      (capitalize breg) 
                      (capitalize ireg) 
                      scale disp; 
        lex lexbuf
      }
  | _ as c 
      { unexpected_char c lexbuf }
  | eof 
      { Printf.printf "Done.\n" }

(* trailer section *)
{
let _ = lex (Lexing.from_channel stdin)
}


